/* $Header: /cvsroot/farplus/StartMgr/StartMgr.cpp,v 1.4 2002/09/05 17:04:33 yole Exp $
   Startup manager plugin for FAR: main source file
   (C) 2001-02 Dmitry Jemerov <yole@yole.ru>
*/

#include <ctype.h>
#include <stdio.h>
#include <atlbase.h>
#include <comdef.h>
#include <shlobj.h>
#include "plugin.hpp"
#include <FARRegistry.h>
#include <FARDialog.h>
#include <FARFile.h>
#include "StartMgr.h"

char PanelTitle [256];
Options Opt;
CStartupManager *gMgr = NULL;

char *TruncRegString (char *Str, int MaxLength)
{
	if(Str)
	{
		char *Root=strchr (Str, '\\');
		if (Root==NULL || Root-Str+5>MaxLength)
			return FarSF::TruncStr(Str,MaxLength);
		Root++;
		int Length=strlen(Str);
		if (Length>MaxLength)
		{
			char *MovePos=Root+Length-MaxLength+3;
			MoveMemory(Root+3,MovePos,strlen(MovePos)+1);
			MoveMemory(Root,"...",3);
		}
	}
	return(Str);
}

// -- CStartupItem -----------------------------------------------------------

CStartupItem::CStartupItem (const char *commandLine)
{
	const char *pathStart = commandLine;
	const char *pathEnd = pathStart;
	if (*pathStart == '\"')
	{
		pathStart++;
		pathEnd = pathStart;
		while (*pathEnd && *pathEnd != '\"')
			pathEnd++;
	}
	else {
		while (*pathEnd && *pathEnd != ' ')
			pathEnd++;
	}
	fStartPath.SetText (pathStart, pathEnd - pathStart);
	
	if (*pathEnd == '\"')
		pathEnd++;
	while (*pathEnd && *pathEnd == ' ')
		pathEnd++;
	fParameters = pathEnd;
}

void CStartupItem::FillPluginPanelItem (PluginPanelItem *panelItem)
{
	memset (panelItem, 0, sizeof (PluginPanelItem));
	strcpy (panelItem->FindData.cFileName, fStartPath);
	if (!fParameters.IsEmpty())
	{
		strncat (panelItem->FindData.cFileName, " ", sizeof (panelItem->FindData.cFileName)-1);
		strncat (panelItem->FindData.cFileName, fParameters, sizeof (panelItem->FindData.cFileName)-1);
	}
	panelItem->UserData = (DWORD) this;
}

// -- CRegistryStartupItem ---------------------------------------------------

CRegistryStartupItem::CRegistryStartupItem (HKEY rootKey, const FarString &regPath, 
					  const FarString &regKey, const FarString &regValue, 
					  int index /* = -1 */, int deleteForbidMessage /* = 0 */)
	: fRootKey (rootKey), 
	  fRegPath (regPath), 
	  fRegKey  (regKey), 
	  fIndex   (index),
	  fDeleteForbidMessage (deleteForbidMessage)
{
    if (regValue [0] == '"')
	{
		// program name in quotation marks, followed by parameters
		int closeQuote = regValue.IndexOf ('"', 1);
		if (closeQuote >= 0)
		{
            fStartPath = regValue.Mid (1, closeQuote-1);
			fParameters = regValue.Mid (closeQuote+1);
			fParameters.Trim();
		}
	}
	else
	{
		if (FarFileInfo::FileExists (regValue))
			fStartPath = regValue;
		else
		{
			int pos = regValue.IndexOf (' ');
			if (pos < 0)
				fStartPath = regValue;
			else
			{
				fStartPath = regValue.Left (pos);
				fParameters = regValue.Mid (pos+1);
				fParameters.Trim();
			}
		}
	}

	fLocation = (fRootKey == HKEY_CURRENT_USER) ? "HKCU\\" : "HKLM\\";
	fLocation += fRegPath;
	fLocation += "\\";
	fLocation += fRegKey;
}


bool CRegistryStartupItem::Delete()
{
	if (fDeleteForbidMessage)
	{
		FarMessage().SimpleMsg (FMSG_WARNING | FMSG_MB_OK, MError, fDeleteForbidMessage, -1);
		return true;
	}
	
	if (fIndex == -1)
		return DeleteKey();

	FarRegistry registry (fRegPath);
	FarString regValue = registry.GetRegStr ("", fRegKey, "", fRootKey);
	FarString newValue;
	int curIndex = 0;
	if (regValue.Length())
	{
		const char *pValue = regValue;
		while (*pValue)
		{
			const char *pValueEnd = pValue+1;
			while (*pValueEnd && *pValueEnd != ',')
				pValueEnd++;
			
			FarString curItem (pValue, pValueEnd - pValue);
			if (curIndex != fIndex)
			{
				newValue += curItem;
				newValue += ",";
			}
			curIndex++;
			pValue = pValueEnd;
			while (*pValue && *pValue == ',')
				pValue++;
		}
	}
	return registry.SetRegKey ("", fRegKey, newValue, fRootKey);
}

bool CRegistryStartupItem::DeleteKey()
{
	HKEY hKey;
	LONG errCode = RegOpenKeyEx (fRootKey, fRegPath, 0, KEY_ALL_ACCESS, &hKey);
	if (errCode != ERROR_SUCCESS)
	{
		SetLastError (errCode);
		return false;
	}

	bool result = true;

	errCode = RegDeleteValue (hKey, fRegKey.ToANSI());
	if (errCode != ERROR_SUCCESS)
	{
		SetLastError (errCode);
		result = false;
	}
	RegCloseKey (hKey);
	return result;
}

// -- CLinkStartupItem -------------------------------------------------------

CLinkStartupItem::CLinkStartupItem (const char *linkPath)
	: fLinkPath (linkPath)
{
	fStartPath = fLinkPath;
	CComPtr<IShellLink> pShellLink;
	pShellLink.CoCreateInstance (CLSID_ShellLink);
	if (pShellLink)
	{
		CComQIPtr<IPersistFile> pFile = pShellLink;
		if (pFile)
		{
			wchar_t wLinkPath [MAX_PATH];
			MultiByteToWideChar (CP_OEMCP, 0, linkPath, -1, wLinkPath, MAX_PATH);
			HRESULT hr = pFile->Load (wLinkPath, STGM_READ);
			if (SUCCEEDED (hr))
			{
				hr = pShellLink->Resolve (NULL, SLR_NO_UI);
				if (SUCCEEDED (hr))
				{
					WIN32_FIND_DATA findData;
					char targetPath [MAX_PATH];
					hr = pShellLink->GetPath (targetPath, MAX_PATH, &findData, 0);
					if (SUCCEEDED (hr))
					{
						fStartPath = targetPath;
						fStartPath = fStartPath.ToOEM();
					}
					hr = pShellLink->GetArguments (targetPath, MAX_PATH);
					if (SUCCEEDED (hr))
					{
						fParameters = targetPath;
						fParameters = fParameters.ToOEM();
					}
				}
			}
		}
	}
}

bool CLinkStartupItem::Delete()
{
	return DeleteFile (fLinkPath) != 0;
}

// -- CRegistryStartupContainer ----------------------------------------------

void CRegistryStartupContainer::Rescan (CStartupItemList &itemList)
{
    FarRegistry registry (fRegPath);
	FarRegistry::ValueIterator valueIter = registry.EnumValues ("", fRootKey);
	FarString keyName, keyValue;
	while (valueIter.NextValue (keyName, keyValue))
	{
        itemList.Add (new CRegistryStartupItem (fRootKey, fRegPath,
			keyName.ToOEM(), keyValue.ToOEM()));
	}
}

void CRegistryStartupContainer::CreateItem (const FarString &itemPath, 
											const FarString &itemName, 
											const FarString &args)
{
	FarString regEntryName = itemName.ToANSI();

	FarRegistry registry (fRegPath);
	// ensure that such entry does not exist
	if (registry.KeyExists ("", regEntryName, fRootKey))
	{
		int curIndex = 1;
		char buf [256];
		do {
			wsprintf (buf, "%s%d", regEntryName.c_str(), curIndex++);
		} while (registry.KeyExists ("", buf, fRootKey));
		regEntryName = buf;
	}

	FarString regValue = itemPath.ToANSI();

	if (!args.IsEmpty())
	{
		FarSF::QuoteSpaceOnly (regValue.GetBuffer (regValue.Length()+2));
		regValue.ReleaseBuffer();
		regValue += " ";
		regValue += args;
	}

	registry.SetRegKey ("", regEntryName, regValue, fRootKey);
}

// -- CRegValueStartupContainer ----------------------------------------------

CRegValueStartupContainer::CRegValueStartupContainer (HKEY rootKey, const char *regKey, 
													  const char *regValue, 
													  int deleteForbidMessage/* =0 */)
	: fRootKey             (rootKey),
	  fRegKey              (regKey),
	  fRegValue            (regValue),
	  fDeleteForbidMessage (deleteForbidMessage)
{
}

void CRegValueStartupContainer::Rescan (CStartupItemList &itemList)
{
	FarRegistry registry (fRegKey);
	FarString regValue = registry.GetRegStr ("", fRegValue, "", fRootKey);

	int curIndex = 0;
	if (regValue.Length())
	{
		const char *pValue = regValue;
		while (*pValue)
		{
			const char *pValueEnd = pValue+1;
			while (*pValueEnd && *pValueEnd != ',')
				pValueEnd++;
			
			FarString cmdLine (pValue, pValueEnd - pValue);
			itemList.Add (new CRegistryStartupItem (fRootKey, fRegKey,
				fRegValue, cmdLine, curIndex++, fDeleteForbidMessage));
			pValue = pValueEnd;
			while (*pValue && *pValue == ',')
				pValue++;
		}
	}
}

// -- CMultiSzTokenizer ------------------------------------------------------

FarString CMultiSzTokenizer::NextToken()
{
    const char *curPtr = fString.c_str() + fPos;
	fPos += strlen (curPtr)+1;
	return FarString (curPtr);
}

// -- CMultiSzStartupContainer -----------------------------------------------

void CMultiSzStartupContainer::Rescan (CStartupItemList &itemList)
{
	FarRegistry registry (fRegKey);
	FarString regValue = registry.GetRegStr ("", fRegValue, "", fRootKey);
    CMultiSzTokenizer tok (regValue);

	while (tok.HasNext())
	{
		itemList.Add (new CRegistryStartupItem (fRootKey, fRegKey,
			fRegValue, tok.NextToken(), -1, fDeleteForbidMessage));
	}
}

// -- CPendingRenameStartupContainer -----------------------------------------

void CPendingRenameStartupContainer::Rescan (CStartupItemList &itemList)
{
	FarRegistry registry (fRegKey);
	FarString regValue = registry.GetRegStr ("", fRegValue, "", fRootKey);
    CMultiSzTokenizer tok (regValue);

	while (tok.HasNext())
	{
		FarString srcPath = tok.NextToken();
		FarString destPath = tok.NextToken();
		// enclose the synthetic command in quotes, so that it doesn't get
		// split into command and parameters by the CRegistryStartupItem
		// constructor
		itemList.Add (new CRegistryStartupItem (fRootKey, fRegKey,
			fRegValue, "\"" + srcPath + " -> " + destPath + "\"",
			-1, fDeleteForbidMessage));
	}
}

// -- CFolderStartupContainer ------------------------------------------------

CFolderStartupContainer::CFolderStartupContainer (HKEY rootKey, const char *regKey, 
												  const char *regValue)
{
	FarRegistry registry (regKey);
	fFolderPath = registry.GetRegStr ("", regValue, "", rootKey).ToOEM();
	if (fFolderPath.Length())
	{
		if (fFolderPath [fFolderPath.Length()-1] != '\\')
			fFolderPath += "\\";
	}
}

void CFolderStartupContainer::Rescan (CStartupItemList &itemList)
{
	if (!fFolderPath.Length())
		return;

	FarFileName searchPath = FarFileName::MakeName (fFolderPath, "*");

	WIN32_FIND_DATA findData;
	HANDLE hFind = FindFirstFile (searchPath, &findData);
	if (hFind != INVALID_HANDLE_VALUE)
	{
		BOOL findResult = TRUE;
		while (findResult)
		{
			if (findData.cFileName [0] != '.' && !(findData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN))
			{
				FarString linkPath = fFolderPath;
				linkPath += findData.cFileName;
				itemList.Add (new CLinkStartupItem (linkPath));
			}
			findResult = FindNextFile (hFind, &findData);
		}
		FindClose (hFind);
	}
}

void CFolderStartupContainer::CreateItem (const FarString &itemPath, 
										  const FarString &itemName, 
										  const FarString &args)
{
	char newLinkPath [260];
	wsprintf (newLinkPath, "%s%s.lnk", fFolderPath.c_str(), itemName.c_str());
	
	// check if filename already exists
	int curIndex = 1;
	while (FarFileInfo::FileExists (newLinkPath))
	{
		wsprintf (newLinkPath, "%s%s%d.lnk", fFolderPath.c_str(), itemName.c_str(), 
			curIndex++);
	}

	BOOL bFileApisAreANSI = AreFileApisANSI();
	if(!bFileApisAreANSI) SetFileApisToANSI();

	CComPtr<IShellLink> pShellLink;
	pShellLink.CoCreateInstance (CLSID_ShellLink);
	if (pShellLink)
	{
		pShellLink->SetPath (itemPath.ToANSI());
		if (!args.IsEmpty())
			pShellLink->SetArguments (args.ToANSI());

		CComQIPtr<IPersistFile> pFile = pShellLink;
		if (pFile)
		{
			wchar_t wLinkPath [MAX_PATH];
			MultiByteToWideChar (CP_OEMCP, 0, newLinkPath, -1, wLinkPath, MAX_PATH);
			pFile->Save (wLinkPath, TRUE);
		}
	}

	if(!bFileApisAreANSI) SetFileApisToOEM();
}

// -- CIniStartupContainer ---------------------------------------------------

void CIniStartupContainer::Rescan (CStartupItemList &itemList)
{
	char iniString [256];
	GetPrivateProfileString ("windows", "load", "", iniString, sizeof (iniString)-1, fIniPath);
	if (lstrlen (iniString) > 0)
		ParseIniLine (itemList, "load", iniString);
	GetPrivateProfileString ("windows", "run", "", iniString, sizeof (iniString)-1, fIniPath);
	if (lstrlen (iniString) > 0)
		ParseIniLine (itemList, "run", iniString);
}

void CIniStartupContainer::ParseIniLine (CStartupItemList &itemList, 
										 const char *key, const char *value)
{
	FarStringTokenizer tok (value);
	int curIndex = 0;
	while (tok.HasNext())
	{
		FarString curValue = tok.NextToken();
        if (!curValue.IsEmpty())
		{
			itemList.Add (new CIniStartupItem (fIniPath, key, curValue, curIndex));
			curIndex++;
		}
	}
}

// -- CIniStartupItem -------------------------------------------------------

bool CIniStartupItem::Delete()
{
	char iniString [256];
	GetPrivateProfileString ("windows", fIniKey, "", iniString, sizeof (iniString)-1,
		fIniPath);

	if (fIndex == 0 && !strchr (iniString, ','))
	{
		// it was the only value => erase the key
		WritePrivateProfileString ("windows", fIniKey, NULL, fIniPath);
		return true;
	}

	// assume the string hasn't changed (fIndex still points to the right item)

	FarString newValue;
	int curIndex = 0;
	bool firstValue = true;

	FarStringTokenizer tok (iniString);
    while (tok.HasNext())
	{
		FarString token = tok.NextToken();
		if (!token.IsEmpty())
		{
			if (curIndex != fIndex)
			{
				if (firstValue)
					firstValue = false;
				else
					newValue += ',';
				newValue += token;
			}
			curIndex++;
		}
	}
	
	WritePrivateProfileString ("windows", fIniKey, newValue, fIniPath);
	return true;
}

// -- CStartupManager --------------------------------------------------------

CStartupManager::CStartupManager()
{
	gMgr = this;

	wsprintf (PanelTitle, " %s ", Far::GetMsg (MStartupMgr));

	fContainerCount = 20;
	fContainers = new CStartupContainer * [fContainerCount];
	fContainers [0] = new CMultiSzStartupContainer (HKEY_LOCAL_MACHINE,
		"SYSTEM\\CurrentControlSet\\Control\\Session Manager", "BootExecute",
		MCannotDeleteMultiSz);
	fContainers [1] = new CPendingRenameStartupContainer (HKEY_LOCAL_MACHINE,
		"SYSTEM\\CurrentControlSet\\Control\\Session Manager", "PendingFileRenameOperations",
		MCannotDeleteMultiSz);
	fContainers [2] = new CRegValueStartupContainer (HKEY_LOCAL_MACHINE, 
		"SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon", "Userinit",
		MCannotDeleteUserinit);
	fContainers [3] = new CRegistryStartupContainer (HKEY_LOCAL_MACHINE,
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\RunOnce");
	fContainers [4] = new CRegistryStartupContainer (HKEY_LOCAL_MACHINE,
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\RunOnceEx");
	fContainers [5] = new CRegValueStartupContainer (HKEY_CURRENT_USER, 
		"Software\\Microsoft\\Windows NT\\CurrentVersion\\Windows", "Run");
	
	fAllUsersRegCtr = new CRegistryStartupContainer (HKEY_LOCAL_MACHINE,
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run");
	fContainers [6] = fAllUsersRegCtr;

	fCurUserRegCtr = new CRegistryStartupContainer (HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\Run");
	fContainers [7] = fCurUserRegCtr;
	
	fAllUsersLinkCtr = new CFolderStartupContainer (HKEY_LOCAL_MACHINE, 
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders", "Common Startup"); 
	fContainers [8] = fAllUsersLinkCtr;

	fContainers [9] = new CFolderStartupContainer (HKEY_LOCAL_MACHINE, 
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders", "Common AltStartup");
	
	fCurUserLinkCtr = new CFolderStartupContainer (HKEY_CURRENT_USER, 
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders", "Startup");
	fContainers [10] = fCurUserLinkCtr;
	
	fContainers [11] = new CRegValueStartupContainer (HKEY_CURRENT_USER, 
		"Software\\Microsoft\\Windows NT\\CurrentVersion\\Windows", "Load");
	fContainers [12] = new CRegistryStartupContainer (HKEY_LOCAL_MACHINE,
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\RunServices");
	fContainers [13] = new CRegistryStartupContainer (HKEY_LOCAL_MACHINE,
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\RunServicesOnce");
	fContainers [14] = new CRegistryStartupContainer (HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\RunServices");
	fContainers [15] = new CRegistryStartupContainer (HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\RunServicesOnce");
	fContainers [16] = new CRegistryStartupContainer (HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\RunOnce");
	fContainers [17] = new CRegistryStartupContainer (HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\RunOnceEx");
	fContainers [18] = new CRegistryStartupContainer (HKEY_LOCAL_MACHINE,
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\policies\\Explorer\\Run");

	fContainers [19] = new CIniStartupContainer ("win.ini");

	Rescan();
}

CStartupManager::~CStartupManager()
{
	if (gMgr == this)
		gMgr = NULL;
	for (int i=0; i<fContainerCount; i++)
		delete fContainers [i];
	delete fContainers;
}

void CStartupManager::Rescan()
{
	fItems.Clear();
	for (int i=0; i<fContainerCount; i++)
		fContainers [i]->Rescan (fItems);
}

void CStartupManager::GetOpenPluginInfo (struct OpenPluginInfo *Info)
{
	Info->StructSize = sizeof (OpenPluginInfo);
	Info->Flags = OPIF_ADDDOTS;
	Info->CurDir = "";
	Info->Format = "Startup";
	
	static struct PanelMode PanelModesArray[10];
	PanelModesArray[4].ColumnTypes="NR";
	PanelModesArray[4].ColumnWidths="0";
	PanelModesArray[4].FullScreen=FALSE;

	Info->PanelModesArray=PanelModesArray;
	Info->PanelModesNumber=sizeof(PanelModesArray)/sizeof(PanelModesArray[0]);
	Info->StartPanelMode='4';
	Info->StartSortMode = SM_UNSORTED;

	Info->PanelTitle = PanelTitle;
}

void CStartupManager::GetFindData (struct PluginPanelItem **ppItem, int *pItemsNumber, 
								   int OpMode)
{
	int itemCount = fItems.Count();
	*pItemsNumber = itemCount;
	*ppItem = (PluginPanelItem *) malloc (itemCount * sizeof (PluginPanelItem));
	for (int i=0; i<itemCount; i++)
	{
		PluginPanelItem &item = (*ppItem) [i];
		fItems [i]->FillPluginPanelItem (&item);
	}
}

int CStartupManager::ProcessKey (int Key, unsigned int ControlState)
{
	if (ControlState == 0 && (Key == VK_F3 || Key == VK_CLEAR))
	{
		struct PanelInfo PInfo;
		FarCtrl farCtrl (this);
		farCtrl.GetPanelInfo (&PInfo);
		if (PInfo.ItemsNumber > 0)
		{
			DWORD userData = PInfo.PanelItems [PInfo.CurrentItem].UserData;
			if (userData != 0)
			{
				CStartupItem *item = (CStartupItem *) userData;
				ViewStartupItem (item);
			}
		}
		return TRUE;
	}
	else if (ControlState == PKF_CONTROL && Key == 'R')
	{
		Rescan();
		
		FarCtrl ctrl (this);
		ctrl.UpdatePanel();
		ctrl.RedrawPanel();
		return TRUE;			
	}
	return FALSE;
}

void AddReadOnlyText (FarDialog *dlg, const char *text, const char *truncText)
{
	if (Far::GetVersion() >= 0x170 && Far::GetBuildNumber() >= 687)
	{
		FarEditCtrl *ctrl = new FarEditCtrl (dlg, text, -1, 60);
		ctrl->SetFlags (DIF_READONLY);
		dlg->AddOwnedControl (ctrl);
	}
	else
		dlg->AddText (truncText);
}

void CStartupManager::ViewStartupItem (CStartupItem *item)
{
	HANDLE hConsole = CreateFile ("CONOUT$", GENERIC_READ, 0, NULL, OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL, NULL);
	CONSOLE_SCREEN_BUFFER_INFO sbi;
	GetConsoleScreenBufferInfo (hConsole, &sbi);
	CloseHandle (hConsole);

	int consoleWidth = sbi.srWindow.Right - sbi.srWindow.Left;

	char strBuf [256];
	
	FarDialog dlg (MStartupItem);
	
	dlg.AddText (MProgram);
	strcpy (strBuf, item->GetProgramPath());
	FarSF::TruncPathStr (strBuf, consoleWidth - 10);
	AddReadOnlyText (&dlg, item->GetProgramPath(), strBuf);

	dlg.AddText (MArguments);
	FarString parameters = item->GetParameters();
	if (parameters.IsEmpty())
		parameters = "<none>";
	AddReadOnlyText (&dlg, item->GetParameters(), parameters);
	
	dlg.AddText (MLocation);
	strcpy (strBuf, item->GetLocation());
	TruncRegString (strBuf, consoleWidth - 10);
	AddReadOnlyText (&dlg, item->GetLocation(), strBuf);
	
	FarSeparatorCtrl sep1 (&dlg);
    FarButtonCtrl btnOk (&dlg, MOk);
	btnOk.SetFlags (DIF_CENTERGROUP);

	dlg.Show();
}

int CStartupManager::PutFiles (struct PluginPanelItem *PanelItem, int ItemsNumber, 
							   int Move, int OpMode)
{
	FarString explicitName;
	FarString arguments;
	CStartupContainer *destCtr = fCurUserLinkCtr;
	if (!(OpMode & OPM_SILENT))
	{
		FarDialog dlg (ItemsNumber == 1 ? MCreateStartupItem : MCreateStartupItems, "StartupCreate");
		FarEditCtrl *edtName = NULL;
		FarEditCtrl *edtArgs = NULL;
		if (ItemsNumber == 1)
		{
			FarString strName = FarSF::PointToName (PanelItem [0].FindData.cFileName);
			int dotPos = strName.IndexOf ('.');
			if (dotPos != -1)
				strName.SetLength (dotPos);

			dlg.AddOwnedControl (new FarTextCtrl (&dlg, MStartupItemName));
			edtName = new FarEditCtrl (&dlg, strName, -1, 50, "StartupManager.ItemName");
			dlg.AddOwnedControl (edtName);
			dlg.AddOwnedControl (new FarTextCtrl (&dlg, MCreateArguments));
			edtArgs = new FarEditCtrl (&dlg, "", -1, 50, "StartupManager.Args");
			dlg.AddOwnedControl (edtArgs);
			dlg.AddOwnedControl (new FarSeparatorCtrl (&dlg));
		}
		
		FarRadioGroup radCreateAs (&dlg, Opt.CreateAs);
		radCreateAs.AddItem (MCurUserLink);
		radCreateAs.AddItem (MAllUsersLink);
		radCreateAs.AddItem (MCurUserReg);
		radCreateAs.AddItem (MAllUsersReg);
		
		FarSeparatorCtrl sep2 (&dlg);
		FarButtonCtrl btnOk (&dlg, MOk);
		FarButtonCtrl btnCancel (&dlg, MCancel, DIF_CENTERGROUP);
		FarControl *retCtrl = dlg.Show();
		if (retCtrl == NULL || retCtrl == &btnCancel)
			return TRUE;
		
		if (edtName)
			explicitName = edtName->GetText();
		if (edtArgs)
			arguments = edtArgs->GetText();
		Opt.CreateAs = radCreateAs.GetSelectedIndex();
		switch (radCreateAs.GetSelectedIndex())
		{
		case 0: destCtr = fCurUserLinkCtr; break;
		case 1: destCtr = fAllUsersLinkCtr; break;
		case 2: destCtr = fCurUserRegCtr; break;
		case 3: destCtr = fAllUsersRegCtr; break;
		}

		FarRegistry reg (Far::GetRootKey(), "\\Yoletir\\StartMgr");
		reg.SetRegKey ("", "CreateAs", Opt.CreateAs);
	}
	
	for (int i=0; i<ItemsNumber; i++)
	{
		char fullName [260];
		GetFullPathName (PanelItem [i].FindData.cFileName, sizeof (fullName)-1, fullName, 
			NULL);
		if (ItemsNumber == 1)
			destCtr->CreateItem (fullName, explicitName, arguments);
		else
		{
			// no explicit name when multiple items
			FarString itemName = FarSF::PointToName (fullName);
			int dotPos = itemName.IndexOf ('.');
			if (dotPos != -1)
				itemName.SetLength (dotPos);

			destCtr->CreateItem (fullName, itemName, FarString());
		}

	}
	Rescan();
	return TRUE;
}

int CStartupManager::DeleteFiles (struct PluginPanelItem *PanelItem, int ItemsNumber, 
								  int OpMode)
{
	if (!(OpMode & OPM_SILENT))
	{
		char confirmBuf [256];
		if (ItemsNumber == 1)
		{
			CStartupItem *startItem = (CStartupItem *) PanelItem [0].UserData;
			wsprintf (confirmBuf, Far::GetMsg (MDeleteOne), 
				FarSF::PointToName (startItem->GetProgramPath()));
		}
		else
			wsprintf (confirmBuf, Far::GetMsg (MDeleteMulti), ItemsNumber);
		
		FarMessage msg;
		msg.AddLine (MDelete);
		msg.AddLine (MDoYouWishToDelete);
		msg.AddLine (confirmBuf);
		msg.AddButton (MDelete);
		msg.AddButton (MCancel);

		if (msg.Show() != 0)
			return FALSE;
	}
	
	int retValue = TRUE;
	for (int i=0; i<ItemsNumber; i++)
	{
		CStartupItem *startItem = (CStartupItem *) PanelItem [i].UserData;
		if (!startItem->Delete())
		{
			FarMessage().ErrMsg (MDeleteError, FMSG_ERRORTYPE);
			retValue = FALSE;
			break;
		}
	}
		
	Rescan();
	return retValue;
}

void CreateStartupItem()
{
	PanelInfo PInfo = FarCtrl().GetPanelInfo();
	if (PInfo.SelectedItemsNumber == 0)
		return;
	
	CStartupManager *pMgr = gMgr;
	bool newManager = false;
	if (!pMgr)
	{
		pMgr = new CStartupManager();
		newManager = true;
	}

	pMgr->PutFiles (PInfo.SelectedItems, PInfo.SelectedItemsNumber, 0, 0);

	if (newManager)
		delete pMgr;
}

void RecreateMenu()
{
	Far::ClearDiskMenu();
	Far::ClearPluginMenu();
	if (Opt.AddToDisksMenu)
		Far::AddDiskMenu (MStartupMgr, Opt.DisksMenuHotkey);
	if (Opt.AddToPluginsMenu)
		Far::AddPluginMenu (MStartupMgr);
}

HANDLE OpenPluginMenu()
{
	FarMenu menu (MStartupMgr);
	menu.AddItem (MOpenStartupManager);
	menu.AddItem (MCreateStartupItem);
	switch (menu.Show())
	{
	case 0:
		return new CStartupManager();
	case 1:
		CreateStartupItem();
		return INVALID_HANDLE_VALUE;
	}
	return INVALID_HANDLE_VALUE;
}

int ConfigStartupManager()
{
	FarDialog dlg (MStartupMgr, "Config");
	FarCheckCtrl     chkDisksMenu (&dlg, MAddToDisksMenu, &Opt.AddToDisksMenu);
	FarFixEditCtrl   edtDisksMenuHotkey (&dlg, &Opt.DisksMenuHotkey, 5, 1);
	                 edtDisksMenuHotkey.SetNextY();
	FarTextCtrl      txtDisksMenuHotkey (&dlg, MDisksMenuHotkey, 7);
	FarCheckCtrl     chkPluginsMenu (&dlg, MAddToPluginsMenu, &Opt.AddToPluginsMenu);
	FarSeparatorCtrl sep (&dlg);
	FarButtonCtrl    btnOk (&dlg, MOk);
	                 btnOk.EnableValidate();
	FarButtonCtrl    btnCancel (&dlg, MCancel, DIF_CENTERGROUP);
	if (dlg.Show() != &btnOk)
		return 0;
	
	FarRegistry registry (Far::GetRootKey(), "\\Yoletir\\StartMgr");
	registry.SetRegKey ("", "AddToDisksMenu", Opt.AddToDisksMenu);
	registry.SetRegKey ("", "DisksMenuHotkey", Opt.DisksMenuHotkey);
	registry.SetRegKey ("", "AddToPluginsMenu", Opt.AddToPluginsMenu);
	RecreateMenu();
	return 1;
}

// -- exported functions -----------------------------------------------------

void WINAPI SetStartupInfo (const struct PluginStartupInfo *Info)
{
	CoInitialize (NULL);
	int sizeOf = sizeof (PluginPanelItem);
	Far::Init (Info);
	
	FarRegistry registry (Far::GetRootKey(), "\\Yoletir\\StartMgr");
	Opt.CreateAs         = registry.GetRegKey ("", "CreateAs", 0);
	Opt.AddToDisksMenu   = registry.GetRegBool ("", "AddToDisksMenu", true);
	Opt.DisksMenuHotkey  = registry.GetRegKey ("", "DisksMenuHotkey", 0);
	Opt.AddToPluginsMenu = registry.GetRegBool ("", "AddToPluginsMenu", true);

	RecreateMenu();
	Far::AddPluginConfig (MStartupMgr);
}

void WINAPI ExitFAR()
{
	Far::Done();
	CoUninitialize();
}

void WINAPI GetPluginInfo (struct PluginInfo *Info)
{
	Far::GetPluginInfo (Info);
}

HANDLE WINAPI OpenPlugin (int OpenFrom, int Item)
{
	switch (OpenFrom)
	{
	case OPEN_DISKMENU:
		return new CStartupManager();
	case OPEN_PLUGINSMENU:
		return OpenPluginMenu();
	}
	return INVALID_HANDLE_VALUE;
}

void WINAPI ClosePlugin (HANDLE hPlugin)
{
	delete (CStartupManager *) hPlugin;
}

void WINAPI GetOpenPluginInfo (HANDLE hPlugin,struct OpenPluginInfo *Info)
{
	CStartupManager *pMgr = (CStartupManager *) hPlugin;
	pMgr->GetOpenPluginInfo (Info);
}

int WINAPI GetFindData (HANDLE hPlugin,struct PluginPanelItem **pPanelItem,
						int *pItemsNumber,int OpMode)
{
	CStartupManager *pMgr = (CStartupManager *) hPlugin;
	pMgr->GetFindData (pPanelItem, pItemsNumber, OpMode);
	return TRUE;
}

void WINAPI FreeFindData (HANDLE hPlugin, struct PluginPanelItem *PanelItem, 
						  int ItemsNumber)
{
	if (PanelItem)
		free (PanelItem);
}

int WINAPI ProcessKey (HANDLE hPlugin, int Key, unsigned int ControlState)
{
	CStartupManager *pMgr = (CStartupManager *) hPlugin;
    return pMgr->ProcessKey (Key, ControlState);
}

int WINAPI PutFiles (HANDLE hPlugin, struct PluginPanelItem *PanelItem,
                     int ItemsNumber, int Move, int OpMode)
{
    CStartupManager *pMgr = (CStartupManager *) hPlugin;
    return pMgr->PutFiles (PanelItem, ItemsNumber, Move, OpMode);
}

int WINAPI DeleteFiles (HANDLE hPlugin, struct PluginPanelItem *PanelItem, 
						int ItemsNumber, int OpMode)
{
	CStartupManager *pMgr = (CStartupManager *) hPlugin;
    return pMgr->DeleteFiles (PanelItem, ItemsNumber, OpMode);
}

int WINAPI Configure (int ItemNumber)
{
	return ConfigStartupManager();
}
