/* $Header: /cvsroot/farplus/StartMgr/StartMgr.h,v 1.3 2002/09/03 20:32:15 yole Exp $
   Startup Manager: main header file
   (C) 2001-02 Dmitry Jemerov <yole@yole.ru>
*/

#ifndef __STARTMGR_H
#define __STARTMGR_H

#include <FARPlus.h>

class CStartupItem
{
protected:
	FarString fStartPath;
	FarString fParameters;

public:
	CStartupItem() {};
	CStartupItem (const char *commandLine);
	virtual ~CStartupItem() {};

	void FillPluginPanelItem (PluginPanelItem *panelItem);
	virtual bool Delete() = 0;

	const char *GetProgramPath()
		{ return fStartPath; }
	const char *GetParameters()
		{ return fParameters; }
	virtual const char *GetLocation() = 0;
};

class CRegistryStartupItem: public CStartupItem
{
private:
	HKEY fRootKey;
	FarString fRegPath;
	FarString fRegKey;
	FarString fLocation;
	int fIndex;
	int fDeleteForbidMessage;

	bool DeleteKey();
	
public:
	CRegistryStartupItem (HKEY rootKey, const FarString &regPath, 
		const FarString &regKey, const FarString &regValue, int index = -1,
		int deleteForbidMessage = 0);

	virtual ~CRegistryStartupItem() {};
	virtual const char *GetLocation()
		{ return fLocation; }
	
	virtual bool Delete();
};

class CLinkStartupItem: public CStartupItem
{
private:
	FarString fLinkPath;

public:
	CLinkStartupItem (const char *linkPath);

	virtual const char *GetLocation()
		{ return fLinkPath; }
	virtual bool Delete();
};

class CIniStartupItem: public CStartupItem
{
private:
	FarString fIniPath;
	FarString fIniKey;
	int fIndex;

public:
	CIniStartupItem (const char *iniPath, const char *iniKey, const char *iniValue, int index)
		: CStartupItem (iniValue), fIniPath (iniPath), fIniKey (iniKey), fIndex (index) {};

	virtual const char *GetLocation()
		{ return fIniPath; }
	virtual bool Delete();
};

typedef FarArray<CStartupItem> CStartupItemList;

class CStartupContainer
{
public:
	virtual ~CStartupContainer() {};

	virtual void Rescan (CStartupItemList &itemList) = 0;
	virtual void CreateItem (const FarString &itemPath, const FarString &itemName,
		const FarString &args) = 0;
};

class CRegistryStartupContainer: public CStartupContainer
{
private:
	HKEY fRootKey;
	FarString fRegPath;

public:
	CRegistryStartupContainer (HKEY rootKey, const char *regPath)
		: fRootKey (rootKey), fRegPath (regPath) {};

	virtual void Rescan (CStartupItemList &itemList);
	virtual void CreateItem (const FarString &itemPath, const FarString &itemName,
		const FarString &args);
};

class CRegValueStartupContainer: public CStartupContainer
{
protected:
	HKEY fRootKey;
	FarString fRegKey;
	FarString fRegValue;
	int fDeleteForbidMessage;

public:
	CRegValueStartupContainer (HKEY rootKey, const char *regKey, 
		const char *regValue, int deleteForbidMessage=0);

	virtual void Rescan (CStartupItemList &itemList);
	virtual void CreateItem (const FarString &itemPath, const FarString &itemName,
		const FarString &args) {};
};

class CFolderStartupContainer: public CStartupContainer
{
private:
	FarString fFolderPath;

public:
	CFolderStartupContainer (HKEY rootKey, const char *regKey, const char *regValue);

	virtual void Rescan (CStartupItemList &itemList);
	virtual void CreateItem (const FarString &itemPath, const FarString &itemName,
		const FarString &args = NULL);
};

class CIniStartupContainer: public CStartupContainer
{
private:
	FarString fIniPath;
	void ParseIniLine (CStartupItemList &itemList, const char *key, 
		const char *value);

public:
	CIniStartupContainer (const char *iniPath)
		: fIniPath (iniPath) {};

	virtual void Rescan (CStartupItemList &itemList);
	virtual void CreateItem (const FarString &itemPath, const FarString &itemName,
		const FarString &args = NULL) {};
};

class CMultiSzStartupContainer: public CRegValueStartupContainer
{
public:
	CMultiSzStartupContainer (HKEY rootKey, const char *regKey, 
		const char *regValue, int deleteForbidMessage = 0)
		: CRegValueStartupContainer (rootKey, regKey, regValue, deleteForbidMessage) {};
	virtual void Rescan (CStartupItemList &itemList);
};

class CPendingRenameStartupContainer: public CRegValueStartupContainer
{
public:
	CPendingRenameStartupContainer (HKEY rootKey, const char *regKey, const char *regValue,
		int deleteForbidMessage = 0)
		: CRegValueStartupContainer (rootKey, regKey, regValue, deleteForbidMessage) {};
	virtual void Rescan (CStartupItemList &itemList);
};

class CMultiSzTokenizer
{
private:
	FarString fString;
	int fPos;

public:
	CMultiSzTokenizer (const FarString &str)
		: fString (str), fPos (0) {};
	bool HasNext()
		{ return fString [fPos] != '\0'; }
	FarString NextToken();

};

class CStartupManager
{
private:
	int fContainerCount;
	CStartupContainer **fContainers;
	CStartupItemList fItems;

	CStartupContainer *fCurUserLinkCtr, *fAllUsersLinkCtr, 
		*fCurUserRegCtr, *fAllUsersRegCtr;

	void Rescan();
	void ViewStartupItem (CStartupItem *item);

public:
	CStartupManager();
	virtual ~CStartupManager();

	void GetOpenPluginInfo (struct OpenPluginInfo *Info);
	void GetFindData (struct PluginPanelItem **ppItem, int *pItemsNumber, int OpMode);
	int ProcessKey (int Key, unsigned int ControlState);
	int PutFiles (struct PluginPanelItem *PanelItem, int ItemsNumber, int Move, int OpMode);
	int DeleteFiles (struct PluginPanelItem *PanelItem, int ItemsNumber, int OpMode);
};

struct Options
{
	int CreateAs;
	bool AddToDisksMenu;
	int DisksMenuHotkey;
	bool AddToPluginsMenu;
};

enum {
	MError,
	MOk,
	MCancel,

	MStartupMgr,       // string in disks menu

	MStartupItem,      // view dialog caption
	MProgram,
	MArguments,
	MLocation,

	MDelete,
	MDoYouWishToDelete,
	MDeleteOne,
	MDeleteMulti,
	MDeleteError,

	MCreateStartupItem,
	MCreateStartupItems,
	MStartupItemName,
	MCreateArguments,
	MCurUserLink,
	MAllUsersLink,
	MCurUserReg,
	MAllUsersReg,

	MOpenStartupManager,

	MAddToDisksMenu,
	MDisksMenuHotkey,
	MAddToPluginsMenu,

	MCannotDeleteMultiSz,
	MCannotDeleteUserinit
};

#endif
